@@ -328,26 +328,6 @@ func TestManPageFormatting(t *testing.T) {
 	// FIXME: Test backspace between two uncombinable characters
 }
 
-func TestToPattern(t *testing.T) {
-	assert.Assert(t, toPattern("") == nil)
-
-	// Test regexp matching
-	assert.Assert(t, toPattern("G.*S").MatchString("GRIIIS"))
-	assert.Assert(t, !toPattern("G.*S").MatchString("gRIIIS"))
-
-	// Test case insensitive regexp matching
-	assert.Assert(t, toPattern("g.*s").MatchString("GRIIIS"))
-	assert.Assert(t, toPattern("g.*s").MatchString("gRIIIS"))
-
-	// Test non-regexp matching
-	assert.Assert(t, toPattern(")G").MatchString(")G"))
-	assert.Assert(t, !toPattern(")G").MatchString(")g"))
-
-	// Test case insensitive non-regexp matching
-	assert.Assert(t, toPattern(")g").MatchString(")G"))
-	assert.Assert(t, toPattern(")g").MatchString(")g"))
-}
-
 func TestScrollToBottomWrapNextToLastLine(t *testing.T) {
 	reader := reader.NewFromTextForTesting("",
 		"first line\nline two will be wrapped\nhere's the last line")
