@@ -4,6 +4,8 @@ import (
 	"fmt"
 
 	log "github.com/sirupsen/logrus"
+	"github.com/walles/moor/v2/internal/linemetadata"
+	"github.com/walles/moor/v2/internal/search"
 	"github.com/walles/moor/v2/internal/textstyles"
 	"github.com/walles/moor/v2/twin"
 )
@@ -23,7 +25,7 @@ func (m PagerModeViewing) drawFooter(statusText string, spinner string) {
 	m.pager.readerLock.Unlock()
 
 	searchHelp := "'/' to search"
-	if len(m.pager.searchString) > 0 {
+	if !m.pager.search.Inactive() {
 		searchHelp = "'n'/'p' to search next/previous"
 	}
 	helpText := "Press 'ESC' / 'q' to exit, " + colonHelp + searchHelp + ", '&' to filter, 'h' for help"
@@ -161,24 +163,27 @@ func (m PagerModeViewing) onRune(char rune) {
 
 	case '/':
 		p.mode = NewPagerModeSearch(p, SearchDirectionForward, p.scrollPosition)
-		p.setTargetLine(nil)
-		p.searchString = ""
-		p.searchPattern = nil
+		p.search.Clear()
+
+		// Searchers want to scan the whole file, start reading as much as we can
+		reallyHigh := linemetadata.IndexMax()
+		p.setTargetLine(&reallyHigh)
 
 	case '?':
 		p.mode = NewPagerModeSearch(p, SearchDirectionBackward, p.scrollPosition)
-		p.setTargetLine(nil)
-		p.searchString = ""
-		p.searchPattern = nil
+		p.search.Clear()
+
+		// Searchers want to scan the whole file, start reading as much as we can
+		reallyHigh := linemetadata.IndexMax()
+		p.setTargetLine(&reallyHigh)
 
 	case '&':
 		if !p.isShowingHelp {
 			// Filtering the help text is not supported. Feel free to work on
 			// that if you feel that's time well spent.
 			p.mode = NewPagerModeFilter(p)
-			p.searchString = ""
-			p.searchPattern = nil
-			p.filterPattern = nil
+			p.search.Clear()
+			p.filter = search.Search{}
 		}
 
 	case 'g':
@@ -186,8 +191,12 @@ func (m PagerModeViewing) onRune(char rune) {
 		p.setTargetLine(nil)
 
 	case ':':
-		p.mode = &PagerModeColonCommand{pager: p}
-		p.setTargetLine(nil)
+		if len(p.readers) > 1 {
+			p.mode = &PagerModeColonCommand{pager: p}
+			p.setTargetLine(nil)
+		} else {
+			p.mode = &PagerModeInfo{Pager: p, Text: "Pass more files on the command line to be able to switch between them."}
+		}
 
 	// Should match the pagermode-not-found.go previous-search-hit bindings
 	case 'n':
