@@ -18,6 +18,7 @@ import (
 	"time"
 
 	"github.com/walles/moor/v2/internal/linemetadata"
+	"github.com/walles/moor/v2/internal/textstyles"
 	"github.com/walles/moor/v2/internal/util"
 
 	"github.com/alecthomas/chroma/v2"
@@ -25,12 +26,18 @@ import (
 	log "github.com/sirupsen/logrus"
 )
 
-// Files larger than this won't be highlighted
-//
-//revive:disable-next-line:var-naming
-const MAX_HIGHLIGHT_SIZE int64 = 1024 * 1024
+// An 1.7MB file took 2s to highlight. The number for this limit is totally
+// negotiable.
+const MAX_HIGHLIGHT_SIZE int64 = 2_000_000
 
-const DEFAULT_PAUSE_AFTER_LINES = 20_000
+// To cap resource usage when not needed, start by reading this many lines into
+// memory. If the user scrolls near the end or starts searching, we'll read
+// more.
+//
+// Ref: https://github.com/walles/moor/issues/296
+const DEFAULT_PAUSE_AFTER_LINES = 50_000
+
+var DisablePlainCachingForBenchmarking = false
 
 type ReaderOptions struct {
 	// Format JSON input
@@ -58,6 +65,13 @@ type Reader interface {
 	// that the returned first line may be different from the requested one.
 	GetLines(firstLine linemetadata.Index, wantedLineCount int) InputLines
 
+	// GetLines gets the indicated lines from the input. The lines will be stored
+	// in the provided preallocated slice to avoid allocations. The line count is
+	// determined by the capacity of the provided slice.
+	//
+	// The return value is the status text for the returned lines.
+	GetLinesPreallocated(firstLine linemetadata.Index, resultLines *[]NumberedLine) string
+
 	// False when paused. Showing the paused line count is confusing, because
 	// the user might think that the number is the total line count, even though
 	// we are not done yet.
@@ -67,6 +81,11 @@ type Reader interface {
 	ShouldShowLineCount() bool
 }
 
+type Line struct {
+	raw            string
+	plainTextCache atomic.Pointer[string] // Use line.Plain() to access this field
+}
+
 // ReaderImpl reads a file into an array of strings.
 //
 // It does the reading in the background, and it returns parts of the read data
@@ -74,7 +93,7 @@ type Reader interface {
 //
 // This package provides query methods for the struct, no peeking!!
 type ReaderImpl struct {
-	sync.Mutex
+	sync.RWMutex
 
 	lines []*Line
 
@@ -126,7 +145,7 @@ type ReaderImpl struct {
 
 // InputLines contains a number of lines from the reader, plus metadata
 type InputLines struct {
-	Lines []*NumberedLine
+	Lines []NumberedLine
 
 	// "monkey.txt: 1-23/45 51%"
 	StatusText string
@@ -199,16 +218,18 @@ func (reader *ReaderImpl) readStream(stream io.Reader, formatter chroma.Formatte
 
 // Pause if we should pause, otherwise not. Pausing means waiting for
 // pauseAfterLinesUpdated to be signalled in SetPauseAfterLines().
-func (reader *ReaderImpl) maybePause() {
+func (reader *ReaderImpl) maybePause() time.Duration {
+	t0 := time.Now()
+
 	for {
-		reader.Lock()
+		reader.RLock()
 		shouldPause := len(reader.lines) >= reader.pauseAfterLines
-		reader.Unlock()
+		reader.RUnlock()
 
 		if !shouldPause {
 			// Not there yet, no pause
 			reader.setPauseStatus(false)
-			return
+			return time.Since(t0)
 		}
 
 		reader.setPauseStatus(true)
@@ -222,15 +243,26 @@ func (reader *ReaderImpl) maybePause() {
 // It is used both during the initial read of the stream until it ends, and
 // while tailing files for changes.
 func (reader *ReaderImpl) consumeLinesFromStream(stream io.Reader) {
+	// This value affects BenchmarkReadLargeFile() performance. Validate changes
+	// like this:
+	//
+	//   go test -benchmem -run='^$' -bench 'BenchmarkReadLargeFile' ./internal/reader
+
+	const linePoolSize = 1000
+
 	reader.preAllocLines()
 
 	inspectionReader := inspectionReader{base: stream}
-	bufioReader := bufio.NewReader(&inspectionReader)
+	bufioReader := bufio.NewReaderSize(&inspectionReader, 64*1024)
 	completeLine := make([]byte, 0)
 
+	linePool := make([]Line, linePoolSize)
+
 	t0 := time.Now()
 	for {
-		reader.maybePause()
+		// Deduct pause times so the final number says how much time we actually
+		// spent reading.
+		t0 = t0.Add(reader.maybePause())
 
 		keepReadingLine := true
 		eof := false
@@ -276,16 +308,21 @@ func (reader *ReaderImpl) consumeLinesFromStream(stream io.Reader) {
 		}
 
 		newLineString := string(completeLine)
-		newLine := NewLine(newLineString)
+		if len(linePool) == 0 {
+			linePool = make([]Line, linePoolSize)
+		}
+		newLine := &linePool[0]
+		linePool = linePool[1:]
+		newLine.raw = newLineString
 
 		reader.Lock()
 		if len(reader.lines) > 0 && !reader.endsWithNewline {
 			// The last line didn't end with a newline, append to it
 			newLineString = reader.lines[len(reader.lines)-1].raw + newLineString
-			newLine = NewLine(newLineString)
-			reader.lines[len(reader.lines)-1] = &newLine
+			newLine = &Line{raw: newLineString}
+			reader.lines[len(reader.lines)-1] = newLine
 		} else {
-			reader.lines = append(reader.lines, &newLine)
+			reader.lines = append(reader.lines, newLine)
 		}
 		reader.endsWithNewline = true
 
@@ -323,9 +360,9 @@ func (reader *ReaderImpl) consumeLinesFromStream(stream io.Reader) {
 }
 
 func (reader *ReaderImpl) tailFile() error {
-	reader.Lock()
+	reader.RLock()
 	fileName := reader.FileName
-	reader.Unlock()
+	reader.RUnlock()
 	if fileName == nil {
 		return nil
 	}
@@ -344,9 +381,9 @@ func (reader *ReaderImpl) tailFile() error {
 			return nil
 		}
 
-		reader.Lock()
+		reader.RLock()
 		bytesCount := reader.bytesCount
-		reader.Unlock()
+		reader.RUnlock()
 
 		if bytesCount == -1 {
 			log.Debugf("Bytes count unknown for %s, stop tailing", *fileName)
@@ -499,7 +536,7 @@ func NewFromTextForTesting(name string, text string) *ReaderImpl {
 	lines := []*Line{}
 	if len(noExternalNewlines) > 0 {
 		for _, lineString := range strings.Split(noExternalNewlines, "\n") {
-			line := NewLine(lineString)
+			line := Line{raw: lineString}
 			lines = append(lines, &line)
 		}
 	}
@@ -647,13 +684,13 @@ func (reader *ReaderImpl) Wait() error {
 	for !reader.HighlightingDone.Load() {
 	}
 
-	reader.Lock()
-	defer reader.Unlock()
+	reader.RLock()
+	defer reader.RUnlock()
 	return reader.Err
 }
 
 func textAsString(reader *ReaderImpl, shouldFormat bool) string {
-	reader.Lock()
+	reader.RLock()
 
 	text := strings.Builder{}
 	for _, line := range reader.lines {
@@ -661,7 +698,7 @@ func textAsString(reader *ReaderImpl, shouldFormat bool) string {
 		text.WriteString("\n")
 	}
 	result := text.String()
-	reader.Unlock()
+	reader.RUnlock()
 
 	var jsonData any
 	err := json.Unmarshal([]byte(result), &jsonData)
@@ -695,17 +732,17 @@ func isXml(text string) bool {
 func highlightFromMemory(reader *ReaderImpl, formatter chroma.Formatter, options ReaderOptions) {
 	// Is the buffer small enough?
 	var byteCount int64
-	reader.Lock()
+	reader.RLock()
 	for _, line := range reader.lines {
 		byteCount += int64(len(line.raw))
 
 		if byteCount > MAX_HIGHLIGHT_SIZE {
 			log.Info("File too large for highlighting: ", byteCount)
-			reader.Unlock()
+			reader.RUnlock()
 			return
 		}
 	}
-	reader.Unlock()
+	reader.RUnlock()
 
 	text := textAsString(reader, options.ShouldFormat)
 
@@ -751,7 +788,7 @@ func highlightFromMemory(reader *ReaderImpl, formatter chroma.Formatter, options
 	reader.setText(*highlighted)
 }
 
-// createStatusUnlocked() assumes that its caller is holding the lock
+// createStatusUnlocked() assumes that its caller is holding the read lock
 func (reader *ReaderImpl) createStatusUnlocked(lastLine linemetadata.Index) string {
 	displayName := ""
 	if reader.DisplayName != nil {
@@ -813,8 +850,8 @@ func (reader *ReaderImpl) AwaitFirstByte() {
 
 // GetLineCount returns the number of lines available for viewing
 func (reader *ReaderImpl) GetLineCount() int {
-	reader.Lock()
-	defer reader.Unlock()
+	reader.RLock()
+	defer reader.RUnlock()
 
 	return len(reader.lines)
 }
@@ -834,79 +871,164 @@ func (reader *ReaderImpl) ShouldShowLineCount() bool {
 	return false
 }
 
+// The index is for error reporting. Set withCache to false to simulate a cache
+// miss for benchmarking.
+func (line *Line) Plain(index linemetadata.Index) string {
+	fromCache := line.plainTextCache.Load()
+	if DisablePlainCachingForBenchmarking {
+		// Simulate a cache miss for benchmarking
+		fromCache = nil
+	}
+	if fromCache != nil {
+		return *fromCache
+	}
+
+	plain := textstyles.StripFormatting(line.raw, index)
+
+	// If this succeeds, all good. If it fails it means some other goroutine
+	// populated the cache before us, which is also fine.
+	_ = line.plainTextCache.CompareAndSwap(nil, &plain)
+
+	return plain
+}
+
 // GetLine gets a line. If the requested line number is out of bounds, nil is returned.
 func (reader *ReaderImpl) GetLine(index linemetadata.Index) *NumberedLine {
-	reader.Lock()
-	defer reader.Unlock()
+	reader.RLock()
 
 	if index.Index() >= reader.pauseAfterLines-DEFAULT_PAUSE_AFTER_LINES/2 {
+		// Switch to the write lock for changing the pause threshold
+		reader.RUnlock()
+		reader.Lock()
+
 		// Getting close(ish) to the pause threshold, bump it up. The Max()
 		// construct is to handle the case when the add overflows.
 		reader.pauseAfterLines = slices.Max([]int{
 			reader.pauseAfterLines + DEFAULT_PAUSE_AFTER_LINES/2,
 			reader.pauseAfterLines})
+
 		select {
 		case reader.pauseAfterLinesUpdated <- true:
 		default:
 			// Default case required for the write to be non-blocking
 		}
+
+		// Back to read lock for the rest of this function
+		reader.Unlock()
+		reader.RLock()
 	}
 
 	if !index.IsWithinLength(len(reader.lines)) {
+		reader.RUnlock()
 		return nil
 	}
+
+	returnLine := reader.lines[index.Index()]
+	reader.RUnlock()
+
 	return &NumberedLine{
 		Index:  index,
 		Number: linemetadata.NumberFromZeroBased(index.Index()),
-		Line:   reader.lines[index.Index()],
+		Line:   returnLine,
 	}
 }
 
+// Given a starting point and a count, return a start and end index that don't
+// exceed maxIndex. On overflow, the requested range will be shifted backwards
+// to fit within maxIndex, and if that's not enough, cut at the end.
+func clipRangeToLength(start linemetadata.Index, wantedCount int, maxIndex int) (int, int) {
+	if wantedCount <= 0 {
+		panic(fmt.Sprintf("wantedCount must be at least 1, was %d", wantedCount))
+	}
+	if maxIndex < 0 {
+		panic(fmt.Sprintf("maxIndex must be at least 0, was %d", maxIndex))
+	}
+
+	first := start.Index()
+
+	// Cap wantedCount to the available length (maxIndex+1).
+	available := maxIndex + 1
+	if wantedCount > available {
+		wantedCount = available
+	}
+
+	// Clamp start so the window fits: start <= maxIndex - (wantedCount - 1)
+	highestStart := maxIndex - (wantedCount - 1)
+	if first > highestStart {
+		first = highestStart
+	}
+	if first < 0 {
+		first = 0
+	}
+
+	last := first + wantedCount - 1
+	if last > maxIndex {
+		last = maxIndex
+	}
+
+	return first, last
+}
+
 // GetLines gets the indicated lines from the input
-//
-//revive:disable-next-line:unexported-return
 func (reader *ReaderImpl) GetLines(firstLine linemetadata.Index, wantedLineCount int) InputLines {
-	reader.Lock()
-	defer reader.Unlock()
-	return reader.getLinesUnlocked(firstLine, wantedLineCount)
-}
+	reader.RLock()
+	lineCount := len(reader.lines)
+	if lineCount == 0 || wantedLineCount == 0 {
+		statusText := reader.createStatusUnlocked(firstLine)
+		reader.RUnlock()
 
-func (reader *ReaderImpl) getLinesUnlocked(firstLine linemetadata.Index, wantedLineCount int) InputLines {
-	if len(reader.lines) == 0 || wantedLineCount == 0 {
 		return InputLines{
-			StatusText: reader.createStatusUnlocked(firstLine),
+			StatusText: statusText,
 		}
 	}
+	reader.RUnlock()
 
-	lastLine := firstLine.NonWrappingAdd(wantedLineCount - 1)
+	firstLineIndex, lastLineIndex := clipRangeToLength(firstLine, wantedLineCount, lineCount-1)
+	wantedLineCount = lastLineIndex - firstLineIndex + 1
 
-	// Prevent reading past the end of the available lines
-	maxLineIndex := *linemetadata.IndexFromLength(len(reader.lines))
-	if lastLine.IsAfter(maxLineIndex) {
-		lastLine = maxLineIndex
+	resultLines := make([]NumberedLine, 0, wantedLineCount)
+	statusText := reader.GetLinesPreallocated(linemetadata.IndexFromZeroBased(firstLineIndex), &resultLines)
 
-		// If one line was requested, then first and last should be exactly the
-		// same, and we would get there by adding zero.
-		firstLine = lastLine.NonWrappingAdd(1 - wantedLineCount)
+	return InputLines{
+		Lines:      resultLines,
+		StatusText: statusText,
+	}
+}
 
-		return reader.getLinesUnlocked(firstLine, firstLine.CountLinesTo(lastLine))
+// GetLines gets the indicated lines from the input. The lines will be stored
+// in the provided preallocated slice to avoid allocations. The line count is
+// determined by the capacity of the provided slice.
+//
+// The return value is the status text for the returned lines.
+func (reader *ReaderImpl) GetLinesPreallocated(firstLine linemetadata.Index, resultLines *[]NumberedLine) string {
+	// Clear the result slice
+	*resultLines = (*resultLines)[:0]
+
+	reader.RLock()
+
+	if len(reader.lines) == 0 || cap(*resultLines) == 0 {
+		statusText := reader.createStatusUnlocked(firstLine)
+		reader.RUnlock()
+
+		return statusText
 	}
 
-	notNumberedReturnLines := reader.lines[firstLine.Index() : lastLine.Index()+1]
-	returnLines := make([]*NumberedLine, 0, len(notNumberedReturnLines))
-	for loopIndex, line := range notNumberedReturnLines {
-		lineIndex := firstLine.NonWrappingAdd(loopIndex)
-		returnLines = append(returnLines, &NumberedLine{
-			Index:  lineIndex,
-			Number: linemetadata.NumberFromZeroBased(lineIndex.Index()),
-			Line:   line,
+	// Prevent reading past the end of the available lines
+	firstLineIndex, lastLineIndex := clipRangeToLength(firstLine, cap(*resultLines), len(reader.lines)-1)
+
+	statusText := reader.createStatusUnlocked(linemetadata.IndexFromZeroBased(lastLineIndex))
+
+	for loopIndex, returnLine := range reader.lines[firstLineIndex : lastLineIndex+1] {
+		*resultLines = append(*resultLines, NumberedLine{
+			Index:  linemetadata.IndexFromZeroBased(firstLineIndex + loopIndex),
+			Number: linemetadata.NumberFromZeroBased(firstLineIndex + loopIndex),
+			Line:   returnLine,
 		})
 	}
 
-	return InputLines{
-		Lines:      returnLines,
-		StatusText: reader.createStatusUnlocked(lastLine),
-	}
+	reader.RUnlock()
+
+	return statusText
 }
 
 func (reader *ReaderImpl) PumpToStdout() {
@@ -963,7 +1085,7 @@ func (reader *ReaderImpl) PumpToStdout() {
 func (reader *ReaderImpl) setText(text string) {
 	lines := []*Line{}
 	for _, lineString := range strings.Split(text, "\n") {
-		line := NewLine(lineString)
+		line := Line{raw: lineString}
 		lines = append(lines, &line)
 	}
 
